"""
The :mod:`evopreprocess.feature_selection` provides a way to select features with evolutionary and nature inspired
algorithms.
"""

from .EvoFeatureSelection import EvoFeatureSelection

__all__ = ['EvoFeatureSelection']



"""
Helper class which evaluates the sampling and transforms the genotype of evolutionary and nature inspired algorithms
from NiaPy to actual dataset.
"""

# Authors: Sašo Karakatič <karakatic@gmail.com>
# License: GNU General Public License v3.0

import math
import random

import numpy as np
from niapy.problems import Problem
from sklearn.base import ClassifierMixin
from sklearn.metrics import mean_squared_error, f1_score
from sklearn.naive_bayes import GaussianNB
from sklearn.utils import _safe_indexing, check_X_y


class SamplingBenchmark(Problem):
    """
    Helper benchmark class for sampling data.

    Parameters
    ----------
    X : {array-like, sparse matrix}, shape (n_samples, n_features)
        Matrix containing the data which have to be sampled.

    y : array-like, shape (n_samples)
        Corresponding label for each instance in X.

    train_indices : array-like, shape (n_samples)
        Corresponding indices for training instances from X.

    valid_indices : array-like, shape (n_samples)
        Corresponding indices for validation instances from X.

    random_seed : int or None, optional (default=1234)
        It used as seed for the random number generator.

    evaluator : classifier or regressor, optional (default=None)
        The classification or regression object from scikit-learn framework.
        If None, the GausianNB for classification is used.
    """

    # _________________0____1_____2______3_______4___
    mapping = np.array([0.5, 0.75, 0.875, 0.9375, 1])

    def __init__(self,
                 X, y,
                 train_indices=None, valid_indices=None,
                 random_seed=1234,
                 evaluator=None):
        X, y = check_X_y(X, y, force_all_finite=False)
        self.X_train, self.X_valid = X[train_indices, :], X[valid_indices, :]
        self.y_train, self.y_valid = y[train_indices], y[valid_indices]

        super().__init__(self.X_train.shape[0] + 5, 0, 1)  # TODO +5 should be tested

        self.evaluator = GaussianNB() if evaluator is None else evaluator
        self.evaluator.random_state = random_seed
        self.metric = f1_score if issubclass(type(self.evaluator), ClassifierMixin) else mean_squared_error

        self.random_seed = random_seed
        random.seed(random_seed)

    def _evaluate(self, sol):
        phenotype = SamplingBenchmark.map_to_phenotype(self.to_phenotype(sol))
        X_sampled = _safe_indexing(self.X_train, phenotype)
        y_sampled = _safe_indexing(self.y_train, phenotype)

        if X_sampled.shape[0] > 0:  # Check if no instances were selected
            cls = self.evaluator.fit(X_sampled, y_sampled)
            y_predicted = cls.predict(self.X_valid)
            acc = self.metric(self.y_valid, y_predicted)
            # used_percentage = len(y_sampled) / len(sol)

            # Check if classifier or regressor
            acc = (1 - acc) if issubclass(type(self.evaluator), ClassifierMixin) else acc
            # return acc + used_percentage
            return acc
        else:
            return math.inf

    def to_phenotype(self, genotype):
        setting = np.cumsum(genotype[-5:])
        appearances = genotype[:-5]

        if setting[-1] == 0:
            return np.digitize(appearances, setting)
        else:
            return np.digitize(appearances, setting / setting[-1])

    # @staticmethod
    # def genotype_to_map(genotype):
    #     return np.digitize(genotype, SamplingBenchmark.mapping)

    @staticmethod
    def map_to_phenotype(mapping):
        return np.repeat(range(len(mapping)), mapping)



"""
Class to perform instance weighting with evolutionary and nature inspired algorithms.
"""

# Authors: Sašo Karakatič <karakatic@gmail.com>
# License: GNU General Public License v3.0

import logging
import sys
import time
from multiprocessing.pool import Pool

import numpy as np
from niapy.algorithms.basic.ga import GeneticAlgorithm
from niapy.task import OptimizationType, Task
from sklearn.base import ClassifierMixin
from sklearn.model_selection import StratifiedKFold, KFold
from sklearn.naive_bayes import GaussianNB
from sklearn.utils import check_random_state

import evopreprocess.utils.EvoSettings as es
from evopreprocess.data_weighting.WeightingBenchmark import WeightingBenchmark

logging.basicConfig()
logger = logging.getLogger('examples')
logger.setLevel('INFO')


class EvoWeighting(object):
    """
    Weight instances from the dataset with evolutionary and nature-inspired methods.

    Parameters
    ----------
    random_seed : int or None, optional (default=None)
        It used as seed by the random number generator.
        If None, the current system time is used for the seed.

    evaluator : classifier or regressor, optional (default=None)
        The classification or regression object from scikit-learn framework.
        If None, the GausianNB for classification is used.

    optimizer : evolutionary or nature-inspired optimization method, optional (default=GeneticAlgorithm)
        The evolutionary or or nature-inspired optimization method from NiaPy framework.

    n_runs : int, optional (default=10)
        The number of runs on each fold. Only the best performing result of all runs is used.

    n_folds : int, optional (default=3)
        The number of folds for cross-validation split into the training and validation sets.

    benchmark : object, optional (default=WeightingBenchmark)
        The benchmark object with mapping and fitness value calculation.

    n_jobs : int, optional (default=None)
        The number of jobs to run in parallel.
        If None, then the number of jobs is set to the number of cores.

    optimizer_settings : dict, optional (default={})
        Custom settings for the optimizer.
    """

    def __init__(self,
                 random_seed=None,
                 evaluator=None,
                 optimizer=GeneticAlgorithm,
                 n_runs=10,
                 n_folds=2,
                 benchmark=WeightingBenchmark,
                 n_jobs=None,
                 optimizer_settings={}):

        self.evaluator = GaussianNB() if evaluator is None else evaluator
        self.random_seed = int(time.time()) if random_seed is None else random_seed
        self.random_state = check_random_state(self.random_seed)
        self.optimizer = optimizer
        self.n_runs = n_runs
        self.n_folds = n_folds
        self.n_jobs = n_jobs
        self.benchmark = benchmark
        self.optimizer_settings = optimizer_settings

    def reweight(self, X, y):
        """Reweight the dataset.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Matrix containing the data which have to be weighted.

        y : array-like, shape (n_samples)
            Corresponding label for each instance in X.

        Returns
        -------
        weights : ndarray, shape (n_samples)
            The corresponding instance weights of `X` set.

        """

        weights = np.empty((len(y), self.n_folds))  # Columns are weights in one run
        weights.fill(np.nan)

        if self.evaluator is ClassifierMixin:
            skf = StratifiedKFold(n_splits=self.n_folds, shuffle=True, random_state=self.random_state)
        else:
            skf = KFold(n_splits=self.n_folds, shuffle=True, random_state=self.random_state)

        mask = []
        evos = []  # Parameters for parallel threaded evolution run

        for train_index, val_index in skf.split(X, y):
            mask.append(train_index)
            for j in range(self.n_runs):
                evos.append(
                    (X, y, train_index, val_index, self.random_seed + j + 1, self.optimizer, self.evaluator,
                     self.benchmark, self.optimizer_settings))

        with Pool(processes=self.n_jobs) as pool:
            results = pool.starmap(EvoWeighting._run, evos)
        weights = EvoWeighting._reduce(mask, results, self.n_runs, self.n_folds, len(y))
        return weights

    @staticmethod
    def _run(X, y, train_index, val_index, random_seed, optimizer, evaluator, benchmark, optimizer_settings):
        opt_settings = es.get_args(optimizer)
        opt_settings.update(optimizer_settings)
        benchm = benchmark(X=X, y=y,
                           train_indices=train_index, valid_indices=val_index,
                           random_seed=random_seed,
                           evaluator=evaluator)
        task = Task(max_evals=opt_settings.pop('max_evals', 1000),
                    optimization_type=OptimizationType.MINIMIZATION,
                    problem=benchm)

        evo = optimizer(seed=random_seed, **opt_settings)
        r = evo.run(task=task)
        if isinstance(r[0], np.ndarray):
            return benchm.to_phenotype(r[0]), r[1]
        else:
            return benchm.to_phenotype(r[0].x), r[1]

    @staticmethod
    def _reduce(mask, results, runs, cv, len_y=10):
        weights = np.full((len_y, cv), np.nan)  # Columns are number of occurrences in one run

        result_list = [results[x:x + runs] for x in range(0, cv * runs, runs)]
        i = 0
        for cv_one in result_list:
            best_fitness = sys.float_info.max
            best_solution = None
            for result_one in cv_one:
                if (best_solution is None) or (best_fitness > result_one[1]):
                    best_solution, best_fitness = result_one[0], result_one[1]

            weights[mask[i], i] = best_solution
            i = i + 1

        weights = np.nanmean(weights, axis=1)

        return weights



"""
Helper class for default setting of some evolutionary and nature inspired NiaPy algorithms.
"""

# Authors: Sašo Karakatič <karakatic@gmail.com>
# License: GNU General Public License v3.0

import niapy.algorithms.basic as nia

population_size = 10
all_kwargs = {'max_evals': 1000}
ga_kwargs = {'population_size': population_size}
bat_kwargs = {'population_size': population_size}
pso_kwargs = {'population_size': population_size}
es_kwargs = {'population_size': population_size}
abc_kwargs = {'population_size': population_size}
de_kwargs = {'population_size': population_size}
fa_kwargs = {'population_size': population_size}
gw_kwargs = {'population_size': population_size}
hs_kwargs = {'population_size': population_size}
fw_kwargs = {'population_size': population_size}

kwargs = {nia.GeneticAlgorithm: ga_kwargs,
          nia.BatAlgorithm: bat_kwargs,
          nia.ParticleSwarmAlgorithm: pso_kwargs,
          nia.EvolutionStrategyMpL: es_kwargs,
          nia.ArtificialBeeColonyAlgorithm: abc_kwargs,
          nia.DifferentialEvolution: de_kwargs,
          nia.FireflyAlgorithm: fa_kwargs,
          nia.GreyWolfOptimizer: gw_kwargs,
          nia.HarmonySearch: hs_kwargs,
          nia.BareBonesFireworksAlgorithm: fw_kwargs}


def get_args(optimizer):
    """
    Method returns dictionary with default setting for provided optimization method.

    Parameters
    ----------
    optimizer : Optimization method for which default settings will be returned.

    Returns
    -------
    args : Dictionary of default settings for the provided optimization method.
    """
    args = {**all_kwargs,
            **kwargs.get(optimizer, {})}
    return args



"""
Class to perform sampling with evolutionary and nature inspired algorithms.
"""

# Authors: Sašo Karakatič <karakatic@gmail.com>
# License: GNU General Public License v3.0

import logging
import sys
import time
from collections import Counter
from multiprocessing import Pool

import numpy as np
from imblearn.base import BaseSampler
from niapy.algorithms import Individual
from niapy.algorithms.basic.ga import GeneticAlgorithm
from niapy.task import Task, OptimizationType
from niapy.util import objects_to_array
from sklearn.base import ClassifierMixin
from sklearn.model_selection import StratifiedKFold, KFold
from sklearn.naive_bayes import GaussianNB
from sklearn.utils import check_random_state, _safe_indexing

from evopreprocess.data_sampling.SamplingBenchmark import SamplingBenchmark
from evopreprocess.utils import EvoSettings as es

logging.basicConfig()
logger = logging.getLogger('examples')
logger.setLevel('INFO')


class EvoSampling(BaseSampler):
    _sampling_type = 'clean-sampling'

    """
    Sample data with evolutionary and nature-inspired methods.

    Parameters
    ----------
    random_seed : int or None, optional (default=None)
        It used as seed by the random number generator.
        If None, the current system time is used for the seed.

    evaluator : classifier or regressor, optional (default=None)
        The classification or regression object from scikit-learn framework.
        If None, the GausianNB for classification is used.

    optimizer : evolutionary or nature-inspired optimization method, optional (default=GeneticAlgorithm)
        The evolutionary or or nature-inspired optimization method from NiaPy framework.

    n_runs : int, optional (default=10)
        The number of runs on each fold. Only the best performing result of all runs is used.

    n_folds : int, optional (default=3)
        The number of folds for cross-validation split into the training and validation sets.

    benchmark : object, optional (default=SamplingBenchmark)
        The benchmark object with mapping and fitness value calculation.

    n_jobs : int, optional (default=None)
        The number of jobs to run in parallel.
        If None, then the number of jobs is set to the number of cores.
    """

    def __init__(self,
                 random_seed=None,
                 evaluator=None,
                 optimizer=GeneticAlgorithm,
                 n_runs=10,
                 n_folds=2,
                 benchmark=SamplingBenchmark,
                 n_jobs=None,
                 optimizer_settings={}):
        super(EvoSampling, self).__init__()

        if optimizer_settings is None:
            optimizer_settings = {}

        self.evaluator = GaussianNB() if evaluator is None else evaluator
        self.random_seed = int(time.time()) if random_seed is None else random_seed
        self.random_state = check_random_state(self.random_seed)
        self.optimizer = optimizer
        self.n_runs = n_runs
        self.n_folds = n_folds
        self.n_jobs = n_jobs
        self.benchmark = benchmark
        self.optimizer_settings = optimizer_settings

    def fit_resample(self, X, y):
        return self._fit_resample(X, y)

    def _fit_resample(self, X, y):
        """Resample the dataset.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Matrix containing the data which have to be sampled.

        y : array-like, shape (n_samples)
            Corresponding label for each sample in X.

        Returns
        -------
        X_resampled : {ndarray, sparse matrix}, shape (n_samples_new, n_features)
            The array containing the resampled data.

        y_resampled : ndarray, shape (n_samples_new,)
            The corresponding label of `X_resampled`
        """

        if self.evaluator is ClassifierMixin:
            skf = StratifiedKFold(n_splits=self.n_folds, shuffle=True, random_state=self.random_state)
        else:
            skf = KFold(n_splits=self.n_folds, shuffle=True, random_state=self.random_state)

        mask = []
        evos = []  # Parameters for parallel threaded evolution run

        for train_index, val_index in skf.split(X, y):
            mask.append(train_index)
            for j in range(self.n_runs):
                evos.append(
                    (X, y, train_index, val_index, self.random_seed + j + 1, self.optimizer, self.evaluator,
                     self.benchmark, self.optimizer_settings))

        with Pool(processes=self.n_jobs) as pool:
            results = pool.starmap(EvoSampling._run, evos)
        occurrences = EvoSampling._reduce(mask, results, self.n_runs, self.n_folds, self.benchmark, len(y))
        phenotype = self.benchmark.map_to_phenotype(occurrences)
        return (_safe_indexing(X, phenotype),
                _safe_indexing(y, phenotype))

    @staticmethod
    def _run(X, y, train_index, val_index, random_seed, optimizer, evaluator, benchmark, optimizer_settings):
        opt_settings = es.get_args(optimizer)
        opt_settings.update(optimizer_settings)
        benchm = benchmark(X=X, y=y,
                           train_indices=train_index, valid_indices=val_index,
                           random_seed=random_seed,
                           evaluator=evaluator)
        task = Task(max_evals=opt_settings.pop('max_evals', 1000),
                    optimization_type=OptimizationType.MINIMIZATION,
                    problem=benchm)

        # if evaluator is ClassifierMixin:
        #     evo = optimizer(seed=random_seed, InitPopFunc=EvoSampling.heuristicInit, **opt_settings)
        # else:
        evo = optimizer(seed=random_seed, **opt_settings)
        r = evo.run(task=task)
        if isinstance(r[0], np.ndarray):
            return benchm.to_phenotype(r[0]), r[1]
        else:
            return benchm.to_phenotype(r[0].x), r[1]

    @staticmethod
    def _reduce(mask, results, runs, cv, benchmark, len_y=10):
        occurrences = np.full((len_y, cv), np.nan)  # Columns are number of occurrences in one run

        result_list = [results[x:x + runs] for x in range(0, cv * runs, runs)]
        i = 0
        for cv_one in result_list:
            best_fitness = sys.float_info.max
            best_solution = None
            for result_one in cv_one:
                if (best_solution is None) or (best_fitness > result_one[1]):
                    best_solution, best_fitness = result_one[0], result_one[1]

            occurrences[mask[i], i] = best_solution
            i = i + 1

        # occurrences = stats.mode(occurrences, axis=1, nan_policy='omit')[0].flatten()
        occurrences = np.ceil(np.nanmean(occurrences, axis=1))

        return occurrences.astype(np.int8)

    @staticmethod
    def heuristicInit(task, NP, rnd, **kwargs):
        target_stats = Counter(task.benchmark.y_train)
        class_perc = 1 - np.array(list(target_stats.values())) / sum(target_stats.values())
        instance_perc = class_perc.take(task.benchmark.y_train)

        pop = np.random.normal(0, 0.25, (NP, task.D))
        pop = pop + instance_perc
        out_of_range = (pop < 0) | (pop > 1)
        while np.any(out_of_range):
            pop[out_of_range] = np.random.normal(0, 0.25, len(pop[out_of_range])) + np.tile(instance_perc, (NP, 1))[
                out_of_range]
            out_of_range = (pop < 0) | (pop > 1)

        pop = objects_to_array([Individual(task=task, rnd=rnd, e=True, x=pop[i]) for i in range(NP)])
        return pop, np.asarray([x.f for x in pop])



"""
Helper class which evaluates the weights of instances and transforms the genotype of evolutionary and nature
inspired algorithms from NiaPy to actual weights.
"""

# Authors: Sašo Karakatič <karakatic@gmail.com>
# License: GNU General Public License v3.0
import random

from niapy.problems import Problem
from sklearn.base import ClassifierMixin
from sklearn.metrics import accuracy_score, mean_squared_error
from sklearn.naive_bayes import GaussianNB
from sklearn.utils import check_X_y


class WeightingBenchmark(Problem):
    """
    Helper benchmark class for weighting instances.

    Parameters
    ----------
    X : {array-like, sparse matrix}, shape (n_samples, n_features)
        Matrix containing the data which have to be weighted.

    y : array-like, shape (n_samples)
        Corresponding label for each instance in X.

    train_indices : array-like, shape (n_samples)
        Corresponding indices for training instances from X.

    valid_indices : array-like, shape (n_samples)
        Corresponding indices for validation instances from X.

    random_seed : int or None, optional (default=1234)
        It used as seed by the random number generator.

    evaluator : classifier or regressor, optional (default=None)
        The classification or regression object from scikit-learn framework.
        If None, the GausianNB for classification is used.
    """

    def __init__(self,
                 X, y,
                 train_indices=None, valid_indices=None,
                 random_seed=1234,
                 evaluator=None):
        X, y = check_X_y(X, y, force_all_finite=False)
        self.X_train, self.X_valid = X[train_indices], X[valid_indices]
        self.y_train, self.y_valid = y[train_indices], y[valid_indices]

        super().__init__(self.X_train.shape[0] + 1, 0, 2)  # TODO: 2 should be tested

        self.evaluator = GaussianNB() if evaluator is None else evaluator
        self.evaluator.random_state = random_seed
        self.metric = accuracy_score if self.evaluator is ClassifierMixin else mean_squared_error

        self.random_seed = random_seed
        random.seed(random_seed)

    def _evaluate(self, sol):
        weights = self.to_phenotype(sol)
        cls = self.evaluator.fit(self.X_train, self.y_train, sample_weight=weights)
        y_predicted = cls.predict(self.X_valid)
        acc = self.metric(self.y_valid, y_predicted)
        acc = (1 - acc) if self.evaluator is ClassifierMixin else acc
        return acc

    def to_phenotype(self, sol):
        max_value = sol[-1] / self.upper[0]
        weights = sol[:-1]
        return weights * max_value



import unittest

import numpy as np

from evopreprocess.data_sampling.SamplingBenchmark import SamplingBenchmark


class MyTestCase(unittest.TestCase):

    def test_phenotype(self):
        raw_gene = [0.123, 0.57, 0, 0.78, 1, 0.7, 0.4, 0.5]
        appearances_gene = [0.1, 0.2, 0.4, 0.6, 0.8]
        result = SamplingBenchmark.to_phenotype(np.concatenate([raw_gene, appearances_gene]))
        expected = np.array([1, 3, 0, 4, 5, 4, 3, 3])
        np.testing.assert_array_equal(result, expected)

    def test_phenotype_all_appearances_0(self):
        raw_gene = [0.123, 0.57, 0, 0.78, 1, 0.7, 0.4, 0.5]
        appearances_gene = [0, 0, 0, 0, 0]
        result = SamplingBenchmark.to_phenotype(np.concatenate([raw_gene, appearances_gene]))
        expected = np.array([5, 5, 5, 5, 5, 5, 5, 5])
        np.testing.assert_array_equal(result, expected)

    def test_map_to_phenotype(self):
        occurances = [2, 0, 1, 3, 0, 4, 0, 5]
        result = SamplingBenchmark.map_to_phenotype(occurances)
        expected = np.array([0, 0, 2, 3, 3, 3, 5, 5, 5, 5, 7, 7, 7, 7, 7])
        np.testing.assert_array_equal(result, expected)


if __name__ == '__main__':
    unittest.main()



"""
The :mod:`evopreprocess.data_weighting` provides a way to weight instances in datasets with evolutionary and
nature inspired algorithms.
"""

from .EvoWeighting import EvoWeighting

__all__ = ['EvoWeighting']



"""
Helper class which evaluates the feature selection and transforms the genotype of evolutionary and nature inspired
algorithms from NiaPy to actual dataset.
"""

# Authors: Sašo Karakatič <karakatic@gmail.com>
# License: GNU General Public License v3.0

import math
import random

import numpy as np
from niapy.problems import Problem
from sklearn.base import ClassifierMixin
from sklearn.metrics import f1_score
from sklearn.metrics import mean_squared_error
from sklearn.naive_bayes import GaussianNB
from sklearn.utils import check_X_y


class FeatureSelectionBenchmark(Problem):
    """
    Helper benchmark class for feature selection.

    Parameters
    ----------
    X : {array-like, sparse matrix}, shape (n_samples, n_features)
        Matrix containing the data which have to have features selected.

    y : array-like, shape (n_samples)
        Corresponding target values for each instance in X.

    train_indices : array-like, shape (n_samples)
        Corresponding indices for training instances from X.

    valid_indices : array-like, shape (n_samples)
        Corresponding indices for validation instances from X.

    random_seed : int or None, optional (default=1234)
        It used as seed by the random number generator.

    evaluator : classifier or regressor, optional (default=None)
        The classification or regression object from scikit-learn framework.
        If None, the GausianNB for classification is used.
    """

    # _________________0_|_1
    mapping = np.array([0.5])

    def __init__(self,
                 X, y,
                 train_indices=None, valid_indices=None,
                 random_seed=1234,
                 evaluator=None,
                 split=None):

        self.split = split
        X, y = check_X_y(X, y, force_all_finite=False)
        self.X_train, self.X_valid = X[train_indices, :], X[valid_indices, :]
        self.y_train, self.y_valid = y[train_indices], y[valid_indices]

        super().__init__(self.X_train.shape[1] + 1, 0, 1)

        self.evaluator = GaussianNB() if evaluator is None else evaluator
        self.evaluator.random_state = random_seed
        self.metric = f1_score if issubclass(type(self.evaluator), ClassifierMixin) else mean_squared_error

        self.random_seed = random_seed
        random.seed(random_seed)

    def _evaluate(self, sol):
        phenotype = self.to_phenotype(sol)
        X_train_new = self.X_train[:, phenotype]
        X_valid_new = self.X_valid[:, phenotype]

        if X_train_new.shape[1] > 0:  # Check if no features were selected
            cls = self.evaluator.fit(X_train_new, self.y_train)
            y_predicted = cls.predict(X_valid_new)
            acc = self.metric(self.y_valid, y_predicted)
            # used_percentage = X_train_new.shape[1] / len(sol)

            # Check if classifier or regressor
            acc = (1 - acc) if issubclass(type(self.evaluator), ClassifierMixin) else acc
            return acc
        else:
            return math.inf

    def to_phenotype(self, genotype):
        if self.split is None:
            s = genotype[-1]
            features = genotype[:-1]
        else:
            s = self.split
            features = genotype
        return features >= s

    @staticmethod
    def genotype_to_map(genotype):
        return np.digitize(genotype, FeatureSelectionBenchmark.mapping)

    @staticmethod
    def map_to_phenotype(mapping):
        return np.where(mapping == 1)[0]



"""``evopreprocess``: Data preprocessing Python toolkit with evolutionary and nature inspired algorithms.
The main problems tackled by the toolkit are **data sampling** (over- and under-sampling data instances
and mixture of both), **instance weighting** and **feature selection**.

Subpackages
-----------
feature_selection
    Module for feature selection with evolutionary and nature inspired algorithms.
data_sampling
    Module for sampling datasets with evolutionary and nature inspired algorithms.
data_weighting
    Module for weighting instances with evolutionary and nature inspired algorithms.
"""

from . import data_sampling
from . import data_weighting
from . import feature_selection

name = 'evopreprocess'
__project__ = 'evopreprocess'
__version__ = '0.4.5'
__all__ = ['feature_selection', 'data_sampling', 'data_weighting']



import logging
import os

import setuptools

PACKAGE_NAME = 'evopreprocess'


def read_package_variable(key, filename='__init__.py'):
    """Read the value of a variable from the package without importing."""
    module_path = os.path.join(PACKAGE_NAME, filename)
    with open(module_path) as module:
        for line in module:
            parts = line.strip().split(' ', 2)
            if parts[:-1] == [key, '=']:
                return parts[-1].strip("'")
    logging.warning("'%s' not found in '%s'", key, module_path)
    return None


with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name=read_package_variable('__project__'),
    version=read_package_variable('__version__'),
    author='Sašo Karakatič',
    author_email='karakatic@gmail.com',
    description='Data Preprocessing with Evolutionary and Nature Inspired Algorithms.',
    license='GPLv3',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/karakatic/EvoPreprocess',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Information Analysis'
    ],
    install_requires=[
        'numpy',
        'scipy',
        'scikit-learn',
        'imbalanced-learn',
        'NiaPy>=2.0.0'
    ],
    keywords=[
        'Evolutionary Algorithms',
        'Nature Inspired Algorithms',
        'Data Sampling',
        'Instance Weighting',
        'Feature Selection',
        'Preprocessing',
        'Machine Learning'
    ]
)



"""
Class to perform feature selection with evolutionary and nature inspired algorithms.
"""

# Authors: Sašo Karakatič <karakatic@gmail.com>
# License: GNU General Public License v3.0


import logging
import sys
import time
from multiprocessing import Pool

import numpy as np
from niapy.algorithms.basic.ga import GeneticAlgorithm
from niapy.task import Task, OptimizationType
from scipy import stats
from sklearn.base import ClassifierMixin
from sklearn.feature_selection._univariate_selection import _BaseFilter
from sklearn.model_selection import StratifiedKFold, KFold
from sklearn.naive_bayes import GaussianNB
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_is_fitted

import evopreprocess.utils.EvoSettings as es
from evopreprocess.feature_selection.FeatureSelectionBenchmark import FeatureSelectionBenchmark

logging.basicConfig()
logger = logging.getLogger('examples')
logger.setLevel('INFO')


class EvoFeatureSelection(_BaseFilter):
    """
    Select features from the dataset with evolutionary and nature-inspired methods.

    Parameters
    ----------
    random_seed : int or None, optional (default=None)
        It used as seed by the random number generator.
        If None, the current system time is used for the seed.

    evaluator : classifier or regressor, optional (default=None)
        The classification or regression object from scikit-learn framework.
        If None, the GausianNB for classification is used.

    optimizer : evolutionary or nature-inspired optimization method, optional (default=GeneticAlgorithm)
        The evolutionary or or nature-inspired optimization method from NiaPy framework.

    n_runs : int, optional (default=3)
        The number of runs on each fold. Only the best performing result of all runs is used.

    n_folds : int, optional (default=3)
        The number of folds for cross-validation split into the training and validation sets.

    benchmark : object, optional (default=FeatureSelectionBenchmark)
        The benchmark object with mapping and fitness value calculation.

    n_jobs : int, optional (default=None)
        The number of jobs to run in parallel.
        If None, then the number of jobs is set to the number of cores.

    optimizer_settings : dict, optional (default={})
        Custom settings for the optimizer.
    """

    def __init__(self,
                 random_seed=None,
                 evaluator=None,
                 optimizer=GeneticAlgorithm,
                 n_runs=10,
                 n_folds=2,
                 benchmark=FeatureSelectionBenchmark,
                 n_jobs=None,
                 optimizer_settings={}):
        # self.score_func = self.select
        super(EvoFeatureSelection, self).__init__(self.select)

        self.evaluator = GaussianNB() if evaluator is None else evaluator
        self.random_seed = int(time.time()) if random_seed is None else random_seed
        self.random_state = check_random_state(self.random_seed)
        self.optimizer = optimizer
        self.n_runs = n_runs
        self.n_folds = n_folds
        self.n_jobs = n_jobs
        self.benchmark = benchmark
        self.optimizer_settings = optimizer_settings

    def _get_support_mask(self):
        check_is_fitted(self, 'scores_')

        return self.scores_ > 0

    def select(self, X, y):
        """Selects features from the dataset.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Matrix containing the data which have to be processed with feature selection.

        y : array-like, shape (n_samples)
            Corresponding label for each instance in X.

        Returns
        -------
        X_new : {ndarray, sparse matrix}, shape (n_samples, n_features_new)
                The array containing the data with selected features.
        """

        if self.evaluator is ClassifierMixin:
            skf = StratifiedKFold(n_splits=self.n_folds, shuffle=True, random_state=self.random_state)
        else:
            skf = KFold(n_splits=self.n_folds, shuffle=True, random_state=self.random_state)

        evos = []  # Parameters for parallel threaded evolution run

        for train_index, val_index in skf.split(X, y):
            for j in range(self.n_runs):
                evos.append(
                    (X, y, train_index, val_index, self.random_seed + j + 1, self.optimizer, self.evaluator,
                     self.benchmark, self.optimizer_settings))

        with Pool(processes=self.n_jobs) as pool:
            results = pool.starmap(EvoFeatureSelection._run, evos)

        return EvoFeatureSelection._reduce(results, self.n_runs, self.n_folds, self.benchmark, X.shape[1])

    @staticmethod
    def _run(X, y, train_index, val_index, random_seed, optimizer, evaluator, benchmark, optimizer_settings):
        opt_settings = es.get_args(optimizer)
        opt_settings.update(optimizer_settings)
        benchm = benchmark(X=X, y=y,
                           train_indices=train_index, valid_indices=val_index,
                           random_seed=random_seed,
                           evaluator=evaluator)
        task = Task(max_evals=opt_settings.pop('max_evals', 1000),
                    optimization_type=OptimizationType.MINIMIZATION,
                    problem=benchm)

        evo = optimizer(seed=random_seed, **opt_settings)
        r = evo.run(task=task)
        if isinstance(r[0], np.ndarray):
            return benchm.to_phenotype(r[0]), r[1]
        else:
            return benchm.to_phenotype(r[0].x), r[1]

    @staticmethod
    def _reduce(results, runs, cv, benchmark, len_y=10):
        features = np.full((len_y, cv), np.nan)  # Columns are number of occurrences in one run

        result_list = [results[x:x + runs] for x in range(0, cv * runs, runs)]
        i = 0
        for cv_one in result_list:
            best_fitness = sys.float_info.max
            best_solution = None
            for result_one in cv_one:
                if (best_solution is None) or (best_fitness > result_one[1]):
                    best_solution, best_fitness = result_one[0], result_one[1]

            features[:, i] = best_solution.astype(int)
            i = i + 1

        features = stats.mode(features, axis=1, nan_policy='omit')[0].flatten()

        if np.sum(features) == 0:
            features[0] = 1

        return features
